@@ -114,6 +114,8 @@ void Http::Init()
    no_cache=false;
 
    use_propfind_now=true;
+   allprop="";
+   allprop_len=0;
 
    retry_after=0;
 
@@ -191,6 +193,7 @@ void Http::ResetRequestData()
    chunk_size=-1;
    chunk_pos=0;
    chunked_trailer=false;
+   propfind=0;
    inflate=0;
    seen_ranges_bytes=false;
 }
@@ -243,6 +246,8 @@ void Http::Send(const char *format,...)
    va_start(va,format);
    xstring& str=xstring::vformat(format,va);
    va_end(va);
+   if(str.length()==0)
+      return;
    LogSend(5,str);
    conn->send_buf->Put(str);
 }
@@ -431,6 +436,22 @@ void Http::DirFile(xstring& path,const xstring& ecwd,const xstring& efile) const
       path.set_substr(base,2,"");
 }
 
+void Http::SendPropfind(const xstring& efile,int depth)
+{
+   SendMethod("PROPFIND",efile);
+   Send("Depth: %d\r\n",depth);
+   if(allprop_len>0)
+   {
+      Send("Content-Type: text/xml\r\n");
+      Send("Content-Length: %d\r\n",allprop_len);
+   }
+}
+void Http::SendPropfindBody()
+{
+   if(allprop_len>0)
+      Send("%s",allprop);
+}
+
 void Http::SendRequest(const char *connection,const char *f)
 {
    xstring efile;
@@ -469,15 +490,6 @@ void Http::SendRequest(const char *connection,const char *f)
    if(mode==CHANGE_DIR && new_cwd && !new_cwd->url)
       add_slash=!new_cwd->is_file;
 
-   const char *allprop=	// PROPFIND request
-      "<?xml version=\"1.0\" ?>"
-      "<propfind xmlns=\"DAV:\">"
-        "<allprop/>"
-      "</propfind>\r\n";
-   if(!QueryBool("use-allprop",hostname))
-      allprop="";
-   int allprop_len=strlen(allprop);
-
    xstring pfile;
    if(proxy && !https)
    {
@@ -537,6 +549,10 @@ void Http::SendRequest(const char *connection,const char *f)
 	 SendMethod(special==HTTP_MOVE?"MOVE":"COPY",efile);
 	 Send("Destination: %s\r\n",special_data.get());
 	 break;
+      case HTTP_PROPFIND:
+	 SendMethod("PROPFIND",efile);
+	 Send("Depth: 1\r\n"); // directory listing required
+	 break;
       case HTTP_NONE:
 	 abort(); // cannot happen
       }
@@ -606,13 +622,7 @@ void Http::SendRequest(const char *connection,const char *f)
       if(mode==CHANGE_DIR)
       {
 	 if(use_propfind_now)
-	 {
-	    SendMethod("PROPFIND",efile);
-	    Send("Depth: 0\r\n"); // no directory listing required
-	    Send("Content-Type: text/xml\r\n");
-	    if(allprop_len>0)
-	       Send("Content-Length: %d\r\n",allprop_len);
-	 }
+	    SendPropfind(efile,0);
 	 else
 	    SendMethod("HEAD",efile);
       }
@@ -631,11 +641,7 @@ void Http::SendRequest(const char *connection,const char *f)
       }
       else if(mode==MP_LIST)
       {
-	 SendMethod("PROPFIND",efile);
-	 Send("Depth: 1\r\n"); // directory listing required
-	 Send("Content-Type: text/xml\r\n");
-	 if(allprop_len>0)
-	    Send("Content-Length: %d\r\n",allprop_len);
+	 SendPropfind(efile,1);
 	 pos=0;
       }
       break;
@@ -648,7 +654,10 @@ void Http::SendRequest(const char *connection,const char *f)
       break;
 
    case ARRAY_INFO:
-      SendMethod("HEAD",efile);
+      if(use_propfind_now)
+	 SendPropfind(efile,0);
+      else
+	 SendMethod("HEAD",efile);
       break;
 
    case RENAME:
@@ -674,10 +683,8 @@ void Http::SendRequest(const char *connection,const char *f)
 	 Send("%s",special_data.get());
       entity_size=NO_SIZE;
    }
-   else if(mode==MP_LIST || (mode==CHANGE_DIR && use_propfind_now))
-   {
-      Send("%s",allprop);
-   }
+   else if(!xstrcmp(last_method,"PROPFIND"))
+      SendPropfindBody();
 
    keep_alive=false;
    chunked=false;
@@ -690,8 +697,18 @@ void Http::SendRequest(const char *connection,const char *f)
    conn->send_buf->SetPos(0);
 }
 
-void Http::SendArrayInfoRequest()
+int Http::SendArrayInfoRequest()
 {
+   // skip to next needed file
+   for(FileInfo *fi=fileset_for_info->curr(); fi; fi=fileset_for_info->next())
+      if(fi->need)
+	 break;
+   if(array_send<fileset_for_info->curr_index())
+      array_send=fileset_for_info->curr_index();
+
+   if(state!=CONNECTED)
+      return 0;
+
    int m=1;
    if(keep_alive && use_head)
    {
@@ -699,17 +716,22 @@ void Http::SendArrayInfoRequest()
       if(m==-1)
 	 m=100;
    }
+   int req_count=0;
    while(array_send-fileset_for_info->curr_index()<m
    && array_send<fileset_for_info->count())
    {
       FileInfo *fi=(*fileset_for_info)[array_send++];
+      if(fi->need==0)
+	 continue;
       xstring *name=&fi->name;
       if(fi->filetype==fi->DIRECTORY && name->last_char()!='/') {
 	 name=&xstring::get_tmp(*name);
 	 name->append('/');
       }
       SendRequest(array_send==fileset_for_info->count()-1 ? 0 : "keep-alive", *name);
+      req_count++;
    }
+   return req_count;
 }
 
 static const char *extract_quoted_header_value(const char *value)
@@ -786,11 +808,17 @@ void Http::HandleHeaderLine(const char *name,const char *value)
    }
    if(!strcasecmp(name,"Last-Modified"))
    {
+      if(!H_20X(status_code))
+	 return;
+
       time_t t=Http::atotm(value);
-      if(opt_date && H_20X(status_code))
+      if(t==ATOTM_ERROR)
+	 return;
+
+      if(opt_date)
 	 *opt_date=t;
 
-      if(mode==ARRAY_INFO && H_20X(status_code))
+      if(mode==ARRAY_INFO)
       {
 	 FileInfo *fi=fileset_for_info->curr();
 	 fi->SetDate(t,0);
@@ -959,6 +987,37 @@ int Http::Do()
    if(Error())
       return m;
 
+   if(propfind && mode==CHANGE_DIR)
+   {
+      if(propfind->Eof())
+      {
+	 const char *b;
+	 int len;
+	 propfind->Get(&b,&len);
+	 FileSet *fs=HttpListInfo::ParseProps(b,len,GetCwd());
+	 if(fs)
+	 {
+	    FileInfo *fi=fs->curr();
+	    if(fi && fi->Has(fi->TYPE))
+	    {
+	       new_cwd->is_file=(fi->filetype!=fi->DIRECTORY);
+	       fi->MakeLongName();
+	       LogNote(9,"%s",fi->longname.get());
+	    }
+	 }
+	 cwd.Set(new_cwd);
+	 cache->SetDirectory(this, "", !cwd.is_file);
+	 state=DONE;
+	 propfind=0;
+	 return MOVED;
+      }
+      if(propfind->Error()) {
+	 SetError(NO_FILE,propfind->ErrorText());
+	 propfind=0;
+	 return MOVED;
+      }
+   }
+
    switch(state)
    {
    case DISCONNECTED:
@@ -969,6 +1028,16 @@ int Http::Do()
 	 state=DONE;
 	 return MOVED;
       }
+      if(mode==ARRAY_INFO)
+      {
+	 // check if we have anything to request
+	 SendArrayInfoRequest();
+	 if(!fileset_for_info->curr())
+	 {
+	    state=DONE;
+	    return MOVED;
+	 }
+      }
       if(!hftp && mode==QUOTE_CMD && !special)
       {
       handle_quote_cmd:
@@ -980,6 +1049,8 @@ int Http::Do()
 	    special=HTTP_COPY;
 	 else if(file && !strncasecmp(file,"MOVE ",5))
 	    special=HTTP_MOVE;
+	 else if(file && !strncasecmp(file,"PROPFIND ",9))
+	    special=HTTP_PROPFIND;
 	 else
 	 {
 	    SetError(NOT_SUPP,0);
@@ -988,7 +1059,9 @@ int Http::Do()
 	 if(special)
 	 {
 	    // METHOD encoded_path data
-	    const char *scan=file+5;
+	    const char *scan=file;
+	    while(*scan && *scan!=' ')
+	       scan++;
 	    while(*scan==' ')
 	       scan++;
 	    file_url.set(https?"https://":"http://");
@@ -1169,13 +1242,17 @@ int Http::Do()
 	 state=DONE;
 	 return MOVED;
       }
-      LogNote(9,_("Sending request..."));
       if(mode==ARRAY_INFO)
       {
-	 SendArrayInfoRequest();
+	 if(SendArrayInfoRequest()==0) {
+	    // nothing to do
+	    state=DONE;
+	    return MOVED;
+	 }
       }
       else
       {
+	 LogNote(9,_("Sending request..."));
 	 SendRequest();
       }
 
@@ -1263,8 +1340,16 @@ int Http::Do()
 		  // we'll have to receive next header
 		  status.set(0);
 		  status_code=0;
-		  if(!fileset_for_info->next())
+		  for(;;)
 		  {
+		     // skip to next needed file
+		     FileInfo *fi=fileset_for_info->next();
+		     if(!fi || fi->need)
+			break;
+		  }
+		  if(!fileset_for_info->curr())
+		  {
+		     // received all requested info.
 		     state=DONE;
 		     return MOVED;
 		  }
@@ -1479,7 +1564,8 @@ int Http::Do()
 		  if(!xstrcmp(last_method,"MKCOL"))
 		     ResMgr::Set("http:use-mkcol",hostname,"no");
 	       }
-	       if(mode==CHANGE_DIR && !xstrcmp(last_method,"PROPFIND"))
+	       if((mode==CHANGE_DIR || mode==ARRAY_INFO)
+	       && !xstrcmp(last_method,"PROPFIND"))
 	       {
 		  use_propfind_now=false;
 		  try_time=0;
@@ -1502,10 +1588,15 @@ int Http::Do()
       }
       if(mode==CHANGE_DIR)
       {
-	 cwd.Set(new_cwd);
-	 cache->SetDirectory(this, "", !cwd.is_file);
-	 state=DONE;
-	 return MOVED;
+	 if(!xstrcmp(last_method,"PROPFIND"))
+	    propfind=new IOBufferFileAccess(this);
+	 else
+	 {
+	    cwd.Set(new_cwd);
+	    cache->SetDirectory(this, "", !cwd.is_file);
+	    state=DONE;
+	    return MOVED;
+	 }
       }
 
       LogNote(9,_("Receiving body..."));
@@ -1522,6 +1613,7 @@ int Http::Do()
       }
       state=RECEIVING_BODY;
       m=MOVED;
+      /*passthrough*/
    case RECEIVING_BODY:
       if(conn->recv_buf->Error() || conn->send_buf->Error())
 	 goto handle_buf_error;
@@ -1941,6 +2033,15 @@ void Http::Reconfig(const char *name)
 
    user_agent=ResMgr::Query("http:user-agent",c);
    use_propfind_now=(use_propfind_now && QueryBool("use-propfind",c));
+
+   allprop=	// PROPFIND request
+      "<?xml version=\"1.0\" ?>"
+      "<propfind xmlns=\"DAV:\">"
+        "<allprop/>"
+      "</propfind>\r\n";
+   if(!QueryBool("use-allprop",c))
+      allprop="";
+   allprop_len=strlen(allprop);
 }
 
 bool Http::SameSiteAs(const FileAccess *fa) const
@@ -2292,7 +2393,7 @@ check_end (const char *p)
    The routine should probably be even more forgiving (as recommended
    by RFC2068), but I do not have the time to write one.
 
-   Return the computed time_t representation, or -1 if all the
+   Return the computed time_t representation, or ATOTM_ERROR if all the
    schemes fail.
 
    Needless to say, what we *really* need here is something like
@@ -2336,7 +2437,7 @@ Http::atotm (const char *time_string)
      whitespace instead of just one (it works that way on all the
      systems I've tested it on).  */
 
-   time_t ut=-1;
+   time_t ut=ATOTM_ERROR;
 
    setlocale(LC_TIME,"C"); // we need english month and week day names
 
